#include <fstream>
#include <sstream>
#include <boost/format.hpp>
#include <boost/algorithm/string.hpp>
#include "textdata.h"

using namespace std;
using namespace boost;

namespace harkio {

// NoiseLocationFile:: 全バージョンの定義
const short NoiseLocationFile::versions[][2] = {{1,0}};

void NoiseLocationFile::underLatestVersion() {

	short size = sizeof(versions) / (sizeof(short) * 2);
	short latest_major = versions[size-1][0];
	short latest_minor = versions[size-1][1];

	if(!(major_version < latest_major ||
			(major_version == latest_major && minor_version <= latest_minor))) {
		std::string error = (boost::format("This version is %1%:%2% larger than %3%:%4%")
		% major_version % minor_version % latest_major % latest_minor).str();
		HARKIO_THROW_ERROR(error);
	}
}

void NoiseLocationFile::read(istream& is) {
	regex regexp;
	smatch match;
	string pattern;
	string line;

	// 初期化
	noise_lists.resize(0);
	is.seekg(0);

	// コメント読み飛ばし
	skipComment(is);

	// ヘッダ
	if (getline(is, line)) {
		CHOMP(line);
		pattern = "Noise-LocationFile";
		regexp = pattern + "\\s*(?:#+.*|)";
		if (!regex_search(line, regexp)) {
			// "ファイルヘッダがNoise-LocationFileではありません。:%1%"
			HARKIO_THROW_ERROR((boost::format("Inappropriate file header:Noise-LocationFile:%1%") % line).str());
		}
	} else {
		HARKIO_THROW_ERROR((boost::format("File loading is incomplete at Noise-LocationFile")).str());
	}

	// コメント読み飛ばし
	skipComment(is);

	// バージョン
	if (getline(is, line)) {
		CHOMP(line);
		pattern = "Version(?:\\s*=\\s*|\\s*)(\\d+)\\.(\\d+)";
		regexp = pattern + "\\s*(?:#+.*|)";
		if (!regex_search(line, match, regexp)) {
			// "バージョンが分かりません:%1%"
			HARKIO_THROW_ERROR((boost::format("Inappropriate:Version:%1%") % line.c_str()).str());
		}
		setVersion(boost::lexical_cast<int>(match.str(1)), boost::lexical_cast<int>(match.str(2)));
	} else {
		HARKIO_THROW_ERROR((boost::format("File loading is incomplete at Version")).str());
	}
	// バージョンチェック
	underLatestVersion();

	// コメント読み飛ばし
	skipComment(is);

	// ノイズの位置数
	int num_of_position;
	if (getline(is, line)) {
		CHOMP(line);
		pattern = "NumOfPosition(?:\\s*=\\s*|\\s*)(\\d+)\\s*";
		regexp = pattern + "\\s*(?:#+.*|)";
		if (!regex_search(line, match, regexp)) {
			// "ノイズの位置数が分かりません:%1%"
			HARKIO_THROW_ERROR((boost::format("Inappropriate NumOfPosition:%1%") % line).str());
		}
		num_of_position = boost::lexical_cast<int>(match.str(1));
	} else {
		HARKIO_THROW_ERROR((boost::format("File loading is incomplete at NumOfPosition")).str());
	}
	// コメント読み飛ばし
	skipComment(is);

	// ノイズ座標
	int count=0;
	string decimal = "[\\-+]?[0-9]+\\.?[0-9]*(?:[eE][\\-+]?[0-9]+|)";
	pattern = (boost::format("Position\\s*[\\-+]?[0-9]+\\s*,\\s*(%1%)\\s*,\\s*(%1%)\\s*,\\s*(%1%)") % decimal).str(); // %% に注意
	regexp = (boost::format("%1%\\s*(?:#+.*|)") % pattern).str();
	while (getline(is, line)) {
		CHOMP(line);
		// コメント読み飛ばし
		skipComment(is);

		noise_lists.resize(count + 1);
		if (regex_search(line, match, regexp)) {
			noise_lists[count].x = boost::lexical_cast<double>(match.str(1));
			noise_lists[count].y = boost::lexical_cast<double>(match.str(2));
			noise_lists[count].z = boost::lexical_cast<double>(match.str(3));
			count++;
		} else {
			// "ノイズの位置情報が分かりません:%1%"
			HARKIO_THROW_ERROR((boost::format("Inappropriate:Position:%1%") % line).str());
		}
	}
	if (num_of_position != count) {
		// "NumOfPositionとノイズの位置数が一致しません。:%1%"
		HARKIO_THROW_ERROR((boost::format("Mismatched number: %1% and NumOfPosition") % line).str());
	}

}

// MICARYLocationFile:: 全バージョンの定義
const short MICARYLocationFile::versions[][2] = {{1,0}};

void MICARYLocationFile::underLatestVersion() {

	short size = sizeof(versions) / (sizeof(short) * 2);
	short latest_major = versions[size-1][0];
	short latest_minor = versions[size-1][1];

	if(!(major_version < latest_major ||
			(major_version == latest_major && minor_version <= latest_minor))) {
		std::string error = (boost::format("This version is %1%:%2% larger than %3%:%4%")
		% major_version % minor_version % latest_major % latest_minor).str();
		HARKIO_THROW_ERROR(error);
	}
}

void MICARYLocationFile::read(istream& is) {

	regex regexp;
	smatch match;
	string pattern;
	string line;

	// 初期化
	mic_lists.resize(0);
	is.seekg(0);

	// コメント読み飛ばし
	skipComment(is);

	// ヘッダ
	if (getline(is, line)) {
		CHOMP(line);
		pattern = "MICARY-LocationFile";
		regexp = pattern + "\\s*(?:#+.*|)";
		if (!regex_search(line, regexp)) {
			// "ファイルヘッダがMICARY-LocationFileではありません。:%1%"
			HARKIO_THROW_ERROR((boost::format("Inappropriate file header:MICARY-LocationFile:%1%") % line).str());
		}
	} else {
		HARKIO_THROW_ERROR((boost::format("File loading is incomplete at MICARY-LocationFile")).str());
	}

	// コメント読み飛ばし
		skipComment(is);

	// バージョン
	if (getline(is, line)) {
		CHOMP(line);
		pattern = "Version(?:\\s*=\\s*|\\s*)(\\d+)\\.(\\d+)";
		regexp = pattern + "\\s*(?:#+.*|)";
		if (!regex_search(line, match, regexp)) {
			// "バージョンが分かりません"
			HARKIO_THROW_ERROR((boost::format("Inappropriate: Version:%1%") % line).str());
		}
		setVersion(boost::lexical_cast<int>(match.str(1)), boost::lexical_cast<int>(match.str(2)));
	} else {
		HARKIO_THROW_ERROR((boost::format("File loading is incomplete at Version")).str());
	}
	// バージョンチェック
	underLatestVersion();

	// コメント読み飛ばし
	skipComment(is);

	// ヘッダ
	int num_of_position;
	if (getline(is, line)) {
		CHOMP(line);
		pattern = "NumOfPosition(?:\\s*=\\s*|\\s*)(\\d+)\\s*";
		regexp = pattern + "\\s*(?:#+.*|)";
		if (!regex_search(line, match, regexp)) {
			// "マイクの位置数が分かりません:%1%"
			HARKIO_THROW_ERROR((boost::format("Inappropriate:NumOfPosition:%1%") % line).str());
		}
		num_of_position = boost::lexical_cast<int>(match.str(1));
	} else {
		HARKIO_THROW_ERROR((boost::format("File loading is incomplete at NumOfPosition")).str());
	}

	// コメント読み飛ばし
	skipComment(is);

	// マイク座標
	int count=0;
	string decimal = "[\\-+]?[0-9]+\\.?[0-9]*(?:[eE][\\-+]?[0-9]+|)";
	pattern = (boost::format("Position\\s*\\d+\\s*,\\s*(%1%)\\s*,\\s*(%1%)\\s*,\\s*(%1%)") % decimal).str(); // %% に注意
	regexp = (boost::format("%1%\\s*(?:#+.*|)") % pattern).str();
	while (getline(is, line)) {
		CHOMP(line);
		// コメント読み飛ばし
		skipComment(is);

		mic_lists.resize(count + 1);
		if (regex_search(line, match, regexp)) {
			mic_lists[count].x = boost::lexical_cast<double>(match.str(1));
			mic_lists[count].y = boost::lexical_cast<double>(match.str(2));
			mic_lists[count].z = boost::lexical_cast<double>(match.str(3));
			count++;
		} else {
			// "マイク位置情報が見つかりません:%1%"
			HARKIO_THROW_ERROR((boost::format("Inappropriate:Positoin:%1%") % line).str());
		}
	}
	if (num_of_position != count) {
		// "NumOfPositionとマイクの位置数が一致しません。:%1%"
		HARKIO_THROW_ERROR((boost::format("Mismatched number: %1% and NumOfPosition") % line).str());
	}
}

// ImpulseResponseListFile:: 全バージョンの定義
const short ImpulseResponseListFile::versions[][2] = {{1,0}};

void ImpulseResponseListFile::underLatestVersion() {

	short size = sizeof(versions) / (sizeof(short) * 2);
	short latest_major = versions[size-1][0];
	short latest_minor = versions[size-1][1];

	if(!(major_version < latest_major ||
			(major_version == latest_major && minor_version <= latest_minor))) {
		std::string error = (boost::format("This version is %1%:%2% larger than %3%:%4%")
		% major_version % minor_version % latest_major % latest_minor).str();
		HARKIO_THROW_ERROR(error);
	}
}

void ImpulseResponseListFile::read(istream& is) {

	regex regexp;
	smatch match;
	string pattern;
	string line;

	// src_list を一番最初にクリアしておく
	src_list.resize(0);
	is.seekg(0);

	// コメント読み飛ばし
	skipComment(is);

	// ヘッダ
	if (getline(is, line)) {
		CHOMP(line);
		pattern = "ImpulseResponse-ListFile";
		regexp = pattern + "(?:\\s*=\\s*|\\s*)(?:#+.*|)";
		if (!regex_search(line, regexp)) {
			// "ヘッダがImpulseResponse-ListFileではありません。:%1%"
			HARKIO_THROW_ERROR((boost::format("Inappropriate file header:ImpulseResponse-ListFile:%1%") % line).str());
		}
	} else {
		// "読み込みを途中で終了しました。"
		HARKIO_THROW_ERROR((boost::format("File loading is incomplete at ImpulseResponse-ListFile")).str());
	}

	// コメント読み飛ばし
	skipComment(is);

	// バージョン
	if (getline(is, line)) {
		CHOMP(line);
		pattern = "Version(?:\\s*=\\s*|\\s*)(\\d+)\\.(\\d+)";
		regexp = pattern + "\\s*(?:#+.*|)";
		if (regex_search(line, match, regexp)) {
			setVersion(lexical_cast<int>(match.str(1)), lexical_cast<int>(match.str(2)));
		} else {
			// "VERSIONが不明です:%1%"
			HARKIO_THROW_ERROR((boost::format("Inappropriate:Version:%1%") % line).str());
		}
		setVersion(boost::lexical_cast<int>(match.str(1)), boost::lexical_cast<int>(match.str(2)));
	} else {
		// "読み込みを途中で終了しました。"
		HARKIO_THROW_ERROR((boost::format("File loading is incomplete at Version")).str());
	}
	// バージョンチェック
	underLatestVersion();

	// コメント読み飛ばし
	skipComment(is);

	// 音源座標、インパルス応答
	string decimal = "[\\-+]?[0-9]+\\.?[0-9]*(?:[eE][\\-+]?[0-9]+|)";
	pattern = (boost::format("\\s*(%1%)\\s*,\\s*(%1%)\\s*,\\s*(%1%)\\s*\%%\%%(.+),\\s*(\\d+)") % decimal).str(); // %% に注意
	regexp = (boost::format("%1%\\s*(?:#+.*|)") % pattern).str();
	int count = 0;
	int index;
	int len;
	int index_end;
	string str;
  
	while (getline(is, line)) {
		CHOMP(line);
		// コメント読み飛ばし
		skipComment(is);
		src_list.resize(count + 1); // 読む度に配列サイズ変更
		if (regex_search(line, match, regexp)) {
			src_list[count].x = boost::lexical_cast<double>(match.str(1));
			src_list[count].y = boost::lexical_cast<double>(match.str(2));
			src_list[count].z = boost::lexical_cast<double>(match.str(3));
			src_list[count].implement_file = toFullPath(reading_file, boost::lexical_cast<string>(match.str(4)));
			src_list[count].cut_start = boost::lexical_cast<int>(match.str(5));

			// 対象区間の終了値を取得する
			index = 0;
			len = 0;	
			src_list[count].cut_end = 0; //初期値設定(HARKTOOL3での初期値)			
			len = line.length();           //１ラインの長さ取得
			index = line.find( "@" );      //@印までの長さ取得

			if(index > 0 )
			{
				index_end = 0;
				index_end = (len-(index+1));                               //@印以降の文字列の長さ取得
				str = line.substr(index+1,index_end);                  //文字列取得
				boost::trim_if(str, boost::is_any_of(" 　"));            //全角半角空白削除
				src_list[count].cut_end = boost::lexical_cast<int>( str ); //文字列を格納
			}

			count++;
		} else {
			// "音源情報のフォーマットが違います:%1%"
			HARKIO_THROW_ERROR((boost::format("Inappropriate line of impulse response's file:%1%") % line).str());
		}
	}

}

// TSPListFile:: 全バージョンの定義
const short TSPListFile::versions[][2] = {{1,0},{1,1}};

void TSPListFile::underLatestVersion() {

	short size = sizeof(versions) / (sizeof(short) * 2);
	short latest_major = versions[size-1][0];
	short latest_minor = versions[size-1][1];

	if(!(major_version < latest_major ||
			(major_version == latest_major && minor_version <= latest_minor))) {
		std::string error = (boost::format("This version is %1%:%2% larger than %3%:%4%")
		% major_version % minor_version % latest_major % latest_minor).str();
		HARKIO_THROW_ERROR(error);
	}
}









/*--------------------------------------------------------------------------------------------------*/
/*                                                                                                                           */
/*           TSPListFile  read                          */ 
/*                                                                                                                           */
/*--------------------------------------------------------------------------------------------------*/
void TSPListFile::read(istream& is) {

	regex regexp;
	smatch match;
	string pattern;
	string line;

	// src_list を一番最初にクリアしておく
	src_list.resize(0);
	is.seekg(0);

	// コメント読み飛ばし
	skipComment(is);

	// ヘッダ
	if (getline(is, line)) {
		CHOMP(line);
		pattern = "TSP-ListFile";
		regexp = pattern + "\\s*(?:#+.*|)";
		if (!regex_search(line, regexp)) {
			// "ヘッダがTSP-ListFileではありません。:%1%"
			HARKIO_THROW_ERROR((boost::format("Inappropriate file header:TSP-ListFile:%1%") % line).str());
		}
	} else {
		HARKIO_THROW_ERROR((boost::format("File loading is incomplete at TSP-ListFile")).str());
	}

	// コメント読み飛ばし
	skipComment(is);

	// バージョン
	if (getline(is, line)) {
		CHOMP(line);
		pattern = "Version(?:\\s*=\\s*|\\s*)(\\d+)\\.(\\d+)";
		regexp = pattern + "\\s*(?:#+.*|)";
		if (regex_search(line, match, regexp)) {
			setVersion(lexical_cast<int>(match.str(1)), lexical_cast<int>(match.str(2)));
		} else {
			// "VERSIONが不明です:%1%"
			HARKIO_THROW_ERROR((boost::format("Inappropriate:Version:%1% ") % line).str());
		}
		setVersion(boost::lexical_cast<int>(match.str(1)), boost::lexical_cast<int>(match.str(2)));
	} else {
		// "読み込みを途中で終了しました。"
		HARKIO_THROW_ERROR((boost::format("File loading is incomplete at Version")).str());
	}
	// バージョンチェック
	underLatestVersion();

	// コメント読み飛ばし
	skipComment(is);

	// 同期加算回数
	if (getline(is, line)) {
		CHOMP(line);
		pattern = "SynchronousAverage(?:\\s*=\\s*|\\s*)(\\d+)";
		regexp = pattern + "\\s*(?:#+.*|)";
		if (regex_search(line, match, regexp)) {
			setSyncAvg(lexical_cast<int>(match.str(1)));
		} else {
			// "同期加算回数が不明です:%1%"
			HARKIO_THROW_ERROR((boost::format("Inappropriate:SynchronousAverage:%1%") % line).str());
		}
	} else {
		// "読み込みを途中で終了しました。"
		HARKIO_THROW_ERROR((boost::format("File loading is incomplete at SynchronousAverage")).str());
	}

	// コメント読み飛ばし
	skipComment(is);

	// オリジナルインパルスファイル
	if (getline(is, line)) {
		CHOMP(line);
		pattern = "OriginalImpulseFltFile(?:\\s*=\\s*|\\s*)(.+)";
		regexp = pattern + "\\s*(?:#+.*|)";
		if (regex_search(line, match, regexp)) {
			setOriginalImpulse(toFullPath(reading_file, match.str(1)));
		} else {
			// "オリジナルインパルスファイルが不明です:%1%"
			HARKIO_THROW_ERROR((boost::format("Inappropriate:OriginalImpulseFltFile:%1%") % line).str());
		}
	} else {
		// "読み込みを途中で終了しました。"
		HARKIO_THROW_ERROR((boost::format("File loading is incomplete at OriginalImpulseFltFile")).str());
	}

	// コメント読み飛ばし
	skipComment(is);

	// TSP オフセット
	if (getline(is, line)) {
		CHOMP(line);
		pattern = "TSPOffset(?:\\s*=\\s*|\\s*)(\\d+)";
		regexp = pattern + "\\s*(?:#+.*|)";
		if (regex_search(line, match, regexp)) {
			setTSPOffset(lexical_cast<int>(match.str(1)));
		} else {
			// "TSPオフセットが不明です:%1%"
			HARKIO_THROW_ERROR((boost::format("Inappropriate:TSPOffset:%1%") % line).str());
		}
	} else {
		// "読み込みを途中で終了しました。"
		HARKIO_THROW_ERROR((boost::format("File loading is incomplete at TSPOffset")).str());
	}

	// コメント読み飛ばし
	skipComment(is);

	// シグナルマックス
	if (getline(is, line)) {
		CHOMP(line);
		pattern = "SignalMax(?:\\s*=\\s*|\\s*)(\\d+)";
		regexp = pattern + "\\s*(?:#+.*|)";
		if (regex_search(line, match, regexp)) {
			setSignalMax(lexical_cast<int>(match.str(1)));
		} else {
			// "シグナルマックスが不明です:%1%"
			HARKIO_THROW_ERROR((boost::format("Inappropriate:SignalMax:%1%") % line).str());
		}
	} else {
		// "読み込みを途中で終了しました。"
		HARKIO_THROW_ERROR((boost::format("File loading is incomplete at SignalMax")).str());
	}

	// コメント読み飛ばし
	skipComment(is);

	// ChannelList オプショナル項目
	std::streampos pos;
	pos = is.tellg();
	if (getline(is, line)) {
		CHOMP(line);
		regexp = "ChannelList(?:\\s*=\\s*|\\s*)([^#]+)\\s*(?:#+.*|)"; // まず、ChannelListかを判別
		if (regex_search(line, match, regexp)) {
			string line2 = match.str(1);
			while (1) {
				regexp = "(^\\d+)(?:\\s*,\\s*|\\s*)(.*)"; // リストの分解
				if (regex_search(line2, match, regexp)) {
					channel_list.push_back(lexical_cast<int>(match.str(1))); // チャンネル番号
				} else {
					// "TSPオフセットが不明です:%1%"
					HARKIO_THROW_ERROR((boost::format("Inappropriate:ChannelList:%1%") % line).str());
				}
				line2 = match.str(2);
				if (line2 == "") {
					break;
				}
			}
		} else {
			is.seekg(pos); // 戻る
		}
	} else {
		// "読み込みを途中で終了しました。"
		HARKIO_THROW_ERROR((boost::format("File loading is incomplete at ChannelList")).str());
	}

	// コメント読み飛ばし
	skipComment(is);

	// 音源座標、インパルス応答
	string decimal = "[\\-+]?[0-9]+\\.?[0-9]*(?:[eE][\\-+]?[0-9]+|)";
	pattern = (boost::format("\\s*(%1%)\\s*,\\s*(%1%)\\s*,\\s*(%1%)\\s*\%%\%%(.+),\\s*(\\d+)") % decimal).str(); // %% に注意
	regexp = (boost::format("%1%\\s*(?:#+.*|)") % pattern).str();
	int count = 0;
	int index;
	int len;
	int index_end;
	string str;
	
	
	while (getline(is, line)) 
	{
		CHOMP(line);
		// コメント読み飛ばし
		skipComment(is);

		src_list.resize(count + 1); // 読む度に配列サイズ変更
		if (regex_search(line, match, regexp)) {
			src_list[count].x = boost::lexical_cast<double>(match.str(1));
			src_list[count].y = boost::lexical_cast<double>(match.str(2));
			src_list[count].z = boost::lexical_cast<double>(match.str(3));
			src_list[count].implement_file = toFullPath(reading_file, boost::lexical_cast<string>(match.str(4)));
			src_list[count].cut_start = boost::lexical_cast<int>(match.str(5));
			
			// 対象区間の終了値を取得する
			index = 0;
			len = 0;	
			src_list[count].cut_end = 0; //初期値設定(HARKTOOL3での初期値)			
			len = line.length();           //１ラインの長さ取得
			index = line.find( "@" );      //@印までの長さ取得

			if(index > 0 )
			{
				index_end = 0;
				index_end = (len-(index+1));                               //@印以降の文字列の長さ取得
				str = line.substr(index+1,index_end);                  //文字列取得
				boost::trim_if(str, boost::is_any_of(" 　"));            //全角半角空白削除
				src_list[count].cut_end = boost::lexical_cast<int>( str ); //文字列を格納
			}

			count++;

		} else {
			// "音源情報のフォーマットが違います:%1%"
			HARKIO_THROW_ERROR((boost::format("Inappropriate line of tsp's file:%1%") % line).str());
		}
	}

}

} // namespace
